/* 
 * IRTOUCH serial touchscreen driver 
 * 
 * Copyright (c) 2011 IRTOUCH SYSTEMS Co,.Ltd 
 */  
 
#include <linux/usb.h>
#include <linux/usb/input.h>
#include <linux/hid.h>
#include <linux/errno.h>   
#include <linux/kernel.h>   
#include <linux/module.h>   
#include <linux/slab.h>   
#include <linux/input.h>   
#include <linux/serio.h>   
#include <linux/init.h>   
#include <asm/uaccess.h>
#include <linux/miscdevice.h>
#include <linux/cdev.h>
#include <linux/delay.h>
  
#define DRIVER_VERSION "v14.3.3.0006"
#define DRIVER_DESC "IRTOUCH serial touchscreen driver v14.3.3.0006" 
#define TOUCH_DEVICE_NAME	"irtouchser"
#define IRTOUCH_MAJOR_NUM	555
#define IRTOUCH_MINOR_NUM 	0
 
#define SET_COMMAND	0x01
#define GET_COMMAND	0x02

#define	GET_CALIB_X	0x2D
#define	SET_CALIB_X	0x2E
#define	GET_CALIB_Y	0x2B
#define	SET_CALIB_Y	0x2C
#define	GET_DEVICE_CONFIG	0x30
#define	SET_DEVICE_CONFIG	0x31

#define SET_DEVICE_COMMAND_MODE	0x71
#define SET_DEVICE_WORK_MODE	0x93

#define CTLCODE_GET_COORDINATE 0xc0
#define CTLCODE_SET_CALIB_PARA_X 0xc1
#define CTLCODE_SET_CALIB_PARA_Y 0xc2
#define CTLCODE_GET_CALIB_PARA_X 0xc3
#define CTLCODE_GET_CALIB_PARA_Y 0xc4
#define CTLCODE_START_CALIB 0xc5
#define CTLCODE_GET_CALIB_STATUS 0xc6
#define CTLCODE_DEVICE_RESET 0xc7

#define IRTOUCH_MAX_LENGTH  64  

#ifndef SERIO_IRTOUCH
# define SERIO_IRTOUCH	0x41
#endif
  
MODULE_AUTHOR("Smart Zhu <smart.ju888@gmail.com>");  
MODULE_DESCRIPTION(DRIVER_DESC);  
MODULE_LICENSE("GPL");  
  
#pragma pack(1)

struct	device_config{
	unsigned char deviceID;
	unsigned char monitorID;
	unsigned char reserved[8];
	unsigned char calibrateStatus;
	unsigned char reserved1[47];
};

struct calib_param{
	long	A00;
	long	A01;
	long	A10;
	unsigned char reserve[28];
};

struct touch_points{
	int		x;
	int		y;
};

struct irtouch_point{
	unsigned char identifier;
	short x;
	short y;
	short z;
	short width;
	short height;
	unsigned char	status;
	unsigned char	reserver;
};

struct irtouch_pkg{
	unsigned char	startBit;
	unsigned char	commandType;
	unsigned char	command;
	unsigned char	pkgLength;
	unsigned char	deviceId;
	struct irtouch_point	points[3];
	unsigned char	actualCounter;
	unsigned char	crc;	
};

#pragma pack()
  
struct irtouch{  
	struct input_dev *dev;  
	struct serio 	*serio;  
	int id;  
	int idx;  
	unsigned char csum;  
	unsigned char data[IRTOUCH_MAX_LENGTH]; 
	struct irtouch_pkg irpkg;
	char phys[32];  
};  

struct device_context{
	char *irtouchName;
	bool startCalib;
	bool isCalibX;
	bool isCalibY;
	bool receiveStatus;
	struct touch_points tpoints;
	struct calib_param calibX;
	struct calib_param calibY;
	struct device_config devConfig;
};

/* 
 * Definitions & global arrays. 
 */  
struct device_context *devContext;
struct irtouch *irtouch;

static void irtouch_translate_points(struct irtouch_point pt, int *x, int *y)
{
	*x = ((devContext->calibX.A01 * pt.y) / 10000) + ((devContext->calibX.A10 * pt.x) / 10000) + devContext->calibX.A00;
	*y = ((devContext->calibY.A01 * pt.y) / 10000) + ((devContext->calibY.A10 * pt.x) / 10000) + devContext->calibY.A00;
}

static int irtouch_build_package(unsigned char command_type, unsigned char command, int device_id, int length, unsigned char *data, unsigned char *out_data)
{
	unsigned char tmp = 0x55;
	unsigned char package[64] = {0};
	int i;
	int j;
	int out_length = 0;

	package[0] = 0xaa;
	package[1] = command_type;
	package[2] = command;
	package[3] = (unsigned char)length;
	package[4] = (unsigned char)device_id;

	if(length > 0)
	{
		length--;
	}
	
	for(i = 0; i < length; i++)
	{
		package[5 + i] = data[i];
	}
	 
	for(j = 0; j < 5 + length; j++)
	{
		tmp = package[j] + tmp;
	}

	package[5 + length] = (unsigned char)tmp;

	out_length = 6 + length;

	memcpy(out_data, package, out_length);

	return out_length;
}

static int send_command(struct serio *serio, unsigned char *inData, int length)
{
	int i = 0;
	int ret = -1;
			
	for(i=0; i<length; i++)
	{
		ret = serio_write(serio,inData[i]);
	}

	return ret;
}

static void irtouch_delay(void)
{
	int iCount = 0;

	while(1)
	{
		if(devContext->receiveStatus || iCount > 50)
		{
			devContext->receiveStatus = false;	
			return;
		}

		mdelay(10);
		iCount++;
	}
}

static void irtouch_get_device_param(struct serio *serio)
{
	unsigned char	inData[58];
	unsigned char	outData[64];
	int status = -1;
	int length = 0;
	
	memset(inData,0,sizeof(inData));
	memset(outData,0,sizeof(outData));

	//set touchscreen to command mode
	inData[0] = 0x01;

	length = irtouch_build_package(SET_COMMAND,SET_DEVICE_COMMAND_MODE,0,2,inData,outData);
	status = send_command(serio,outData,length);
	if(status == -1)
	{
		printk(KERN_ALERT "Set command mode failed.");
	}
	irtouch_delay();

	
	memset(inData,0,sizeof(inData));
	memset(outData,0,sizeof(outData));

	inData[0] = 0xF4;

	length = irtouch_build_package(SET_COMMAND,SET_DEVICE_WORK_MODE,0,2,inData,outData);
	status = send_command(serio,outData,length);
	if(status == -1)
	{
		printk(KERN_ALERT "Set device to work status failed.");
	}
	irtouch_delay();

	//get device configuration
	memset(inData,0,sizeof(inData));
	memset(outData,0,sizeof(outData));

	length = irtouch_build_package(GET_COMMAND,GET_DEVICE_CONFIG,1,1,inData,outData);
	status = send_command(serio,outData,length);
	if(status == -1)
	{
		printk(KERN_ALERT "Get device configuration failed.");
	}
	irtouch_delay();
	memcpy(&devContext->devConfig,&irtouch->data[5],sizeof(struct device_config));

	//get calib x parameter
	memset(inData,0,sizeof(inData));
	memset(outData,0,sizeof(outData));

	length = irtouch_build_package(GET_COMMAND,GET_CALIB_X,0,1,inData,outData);
	status = send_command(serio,outData,length);
	if(status == -1)
	{
		printk(KERN_ALERT "Get calib X param failed.");
	}
	irtouch_delay();
	memcpy(&devContext->calibX,&irtouch->data[5],sizeof(struct calib_param));

	//get calib y parameter
	memset(inData,0,sizeof(inData));
	memset(outData,0,sizeof(outData));

	length = irtouch_build_package(GET_COMMAND,GET_CALIB_Y,0,1,inData,outData);
	status = send_command(serio,outData,length);
	if(status == -1)
	{
		printk(KERN_ALERT "Get calib Y param failed.");
	}
	irtouch_delay();
	memcpy(&devContext->calibY,&irtouch->data[5],sizeof(struct calib_param));

	//set touchscreen to work mode
	memset(inData,0,sizeof(inData));
	memset(outData,0,sizeof(outData));

	length = irtouch_build_package(SET_COMMAND,SET_DEVICE_COMMAND_MODE,0,2,inData,outData);	
	status = send_command(serio,outData,length);
	if(status == -1)
	{
		printk(KERN_ALERT "Set to work mode failed.");
	}
	irtouch_delay();
}

static void irtouch_set_calib_param(struct serio *serio, bool resetFlg)
{
	unsigned char	inData[58];
	unsigned char	outData[64];
	int status = -1;
	int length = 0;
	
	//set touchscreen to command mode
	memset(inData,0,sizeof(inData));
	memset(outData,0,sizeof(outData));

	inData[0] = 0x01;

	length = irtouch_build_package(SET_COMMAND,SET_DEVICE_COMMAND_MODE,0,2,inData,outData);
	status = send_command(serio,outData,length);
	if(status == -1)
	{
		printk(KERN_ALERT "Set to command mode failed.");
	}
	irtouch_delay();

	//Set device configuration
	memset(inData,0,sizeof(inData));
	memset(outData,0,sizeof(outData));

	if(resetFlg)
	{
		devContext->devConfig.calibrateStatus = 0;
		memcpy(&inData,&devContext->devConfig,58);

		length = irtouch_build_package(SET_COMMAND,SET_DEVICE_CONFIG,1,59,inData,outData);
		status = send_command(serio,outData,length);
		if(status == -1)
		{
			printk(KERN_ALERT "Set device configuration failed.");
		}
		irtouch_delay();
	}
	else
	{
		devContext->devConfig.calibrateStatus = 1;
		memcpy(&inData,&devContext->devConfig,58);

		length = irtouch_build_package(SET_COMMAND,SET_DEVICE_CONFIG,1,59,inData,outData);
		status = send_command(serio,outData,length);
		if(status == -1)
		{
			printk(KERN_ALERT "Set device configuration failed.");
		}
		irtouch_delay();

		//Set calib X param
		memset(inData,0,sizeof(inData));
		memset(outData,0,sizeof(outData));
		memcpy(&inData,&devContext->calibX,40);

		length = irtouch_build_package(SET_COMMAND,SET_CALIB_X,0,59,inData,outData);
		status = send_command(serio,outData,length);
		if(status == -1)
		{
			printk(KERN_ALERT "Set calib X param failed.");
		}
		irtouch_delay();

		//Set calib Y param
		memset(inData,0,sizeof(inData));
		memset(outData,0,sizeof(outData));
		memcpy(&inData,&devContext->calibY,40);

		length = irtouch_build_package(SET_COMMAND,SET_CALIB_Y,0,59,inData,outData);
		status = send_command(serio,outData,length);
		if(status == -1)
		{
			printk(KERN_ALERT "Set calib Y param failed.");
		}
		irtouch_delay();
	}

	//set touchscreen to work mode
	memset(inData,0,sizeof(inData));
	memset(outData,0,sizeof(outData));

	length = irtouch_build_package(SET_COMMAND,SET_DEVICE_COMMAND_MODE,0,2,inData,outData);	
	status = send_command(serio,outData,length);
	if(status == -1)
	{
		printk(KERN_ALERT "Set to work mode failed.");
	}
	irtouch_delay();
}
  
static void irtouch_send_report(struct irtouch* irtouch)  
{  
	int i = 0;
	int x = 0;
	int y = 0;
	struct input_dev *dev = irtouch->dev;  
	
	memcpy(&irtouch->irpkg,&irtouch->data[0],46);
	
	//send dual-touch data
	for(i = 0; i < (irtouch->irpkg.actualCounter & 0x1F); i++)
	{
		if(irtouch->irpkg.points[i].status == 0x07)
		{
			x = irtouch->irpkg.points[i].x;
			y = irtouch->irpkg.points[i].y;

			if(devContext->startCalib && (irtouch->irpkg.points[i].identifier == 0))
			{
				devContext->tpoints.x = x;
				devContext->tpoints.y = y;
			}

			if(devContext->devConfig.calibrateStatus && !devContext->startCalib)
			{
				irtouch_translate_points(irtouch->irpkg.points[i],&x,&y);
			}

			//Edge of the acceleration;
			if(x < 2000 && x > 500)
			{
				x = x - 500;			
			}

			else if(x > 31000 && x < 32367)
			{
				x = x + 500;
			}

			if(y < 2000 && y > 500)

			{
				y = y - 500;			
			}
			else if(y > 31000 && y < 32267)
			{

				y = y + 500;
			}
			
			input_report_abs(dev,ABS_MT_TRACKING_ID,irtouch->irpkg.points[i].identifier);
			input_report_abs(dev,ABS_MT_POSITION_X, x);
			input_report_abs(dev,ABS_MT_POSITION_Y, y);
			input_report_abs(dev,ABS_MT_TOUCH_MAJOR, max(x,y));
			input_report_abs(dev,ABS_MT_TOUCH_MINOR, min(x,y));
		}
		else
		{
			input_report_abs(dev,ABS_MT_TOUCH_MAJOR, 0);
		}
		input_mt_sync(dev);
	}

	input_sync(dev);
}  

void irtouch_process_data(struct irtouch *irtouch, unsigned char data)
{	
	irtouch->data[irtouch->idx] = data;  
	
	switch(irtouch->idx){
		case 0:
			if(data != 0xaa)
			{
				irtouch->idx = 0;
				return;			
			}
			break;
		case 1:
			if(data != 0x10)
			{
				irtouch->idx = 0;
				return;
			}
			break;
	}
				
	if(irtouch->data[0] == 0xaa)
	{
		if(irtouch->idx == (irtouch->data[3] + 4))
		{	
			unsigned char CRC  = 0x55;
			int j = 0;

			for(j = 0; j < irtouch->idx; j++)
			{
				CRC = CRC + irtouch->data[j];
			}

			if(CRC == irtouch->data[irtouch->idx])
			{	
				if(irtouch->data[2] == 0x65) //touch report;
				{					
					irtouch_send_report(irtouch);
				}
				else //command report;
				{
					devContext->receiveStatus = true;
				}
			}

			irtouch->idx = 0;
			return;
		}
		else if(irtouch->data[2] == 0x65 && irtouch->idx > 45)
		{
			irtouch->idx = 0;
			return;
		}
	}
	else
	{
		irtouch->idx = 0;
		return;		
	}
	
	irtouch->idx++;
}
 
static irqreturn_t irtouch_interrupt(struct serio *serio, unsigned char data, unsigned int flags)  
{  
	struct irtouch* irtouch = serio_get_drvdata(serio);

	switch(irtouch->id){  
	case 0: 
		irtouch_process_data(irtouch, data);  
		break;
	}  

	return IRQ_HANDLED;  
}  

static long irtouch_ioctl(struct file * filp, unsigned int ctl_code, unsigned long ctl_param)
{
	unsigned char value = 0;
	int ret = -1;
	
	switch (ctl_code)
	{
		case CTLCODE_START_CALIB:				
			ret = copy_from_user(&value, (unsigned char *)ctl_param, sizeof(unsigned char));
			if(ret == 0)
			{
				if(value == 0x01)
				{
					devContext->startCalib = true;
				}
				else
				{
					devContext->startCalib = false;
				}
			}	
			
			break;

		case CTLCODE_GET_COORDINATE:
			ret = copy_to_user((struct touch_points *)ctl_param, &devContext->tpoints, sizeof(struct touch_points));
			if(ret != 0)
			{
				err("IRTOUCH:    %s <CTLCODE_GET_COORDINATE>copy_to_user failed!",__func__);
			}
			break;

		case CTLCODE_GET_CALIB_STATUS:
			{
				int status = devContext->devConfig.calibrateStatus;
				ret = copy_to_user((int *)ctl_param, &status, sizeof(int));
				if(ret != 0)
				{
					err("IRTOUCH:    %s <CTLCODE_GET_CALIB_STATUS>copy_to_user failed!",__func__);
				}
				break;
			}
	
		case CTLCODE_SET_CALIB_PARA_X:
			ret = copy_from_user(&devContext->calibX, (struct calib_param *)ctl_param, sizeof(struct calib_param));
			if(ret != 0)
			{
				err("IRTOUCH:    %s <CTLCODE_SET_CALIB_PARA_X>copy_to_user failed!",__func__);
			}
			else
			{
				devContext->isCalibX = true;
			}
			break;

		case CTLCODE_SET_CALIB_PARA_Y:
			ret = copy_from_user(&devContext->calibY, (struct calib_param *)ctl_param, sizeof(struct calib_param));
			if(ret != 0)
			{
				err("IRTOUCH:    %s <CTLCODE_SET_CALIB_PARA_Y>copy_to_user failed!",__func__);
			}
			else
			{
				devContext->isCalibY = true;
			}
			break;

		case CTLCODE_GET_CALIB_PARA_X:
			ret = copy_to_user((struct calib_param *)ctl_param, &devContext->calibX, sizeof(struct calib_param));
			if(ret != 0)
			{
				err("IRTOUCH:    %s <CTLCODE_GET_CALIB_PARA>copy_to_user failed!",__func__);
			}
			break;

		case CTLCODE_GET_CALIB_PARA_Y:
			ret = copy_to_user((struct calib_param *)ctl_param, &devContext->calibY, sizeof(struct calib_param));
			if(ret != 0)
			{
				err("IRTOUCH:    %s <CTLCODE_GET_CALIB_PARA>copy_to_user failed!",__func__);
			}
			break;

		case CTLCODE_DEVICE_RESET:
			ret = copy_from_user(&value, (unsigned char *)ctl_param, sizeof(unsigned char));
			if(ret == 0)
			{
				if(value == 0x01)
				{
					irtouch_set_calib_param(irtouch->serio,true);
				}
			}
			break;
	}

	if(devContext->isCalibX && devContext->isCalibY)
	{
		devContext->isCalibX = false;
		devContext->isCalibY = false;

		irtouch_set_calib_param(irtouch->serio,false);
	}

	return ret;
}
  
static struct class *irser_class;
dev_t dev = 0;
struct cdev cdev;

static struct file_operations ts_fops = {
	owner:	THIS_MODULE,
	unlocked_ioctl:irtouch_ioctl,
};

/* 
 * irtouch_disconnect() is the opposite of irtouch_connect() 
 */    
static void irtouch_disconnect(struct serio *serio)  
{  
	struct irtouch* irtouch = serio_get_drvdata(serio);  

	cdev_del(&cdev);
	unregister_chrdev_region(dev,1);

	device_destroy(irser_class,MKDEV(IRTOUCH_MAJOR_NUM,0));
	class_destroy(irser_class);

	input_unregister_device(irtouch->dev);  
	serio_close(serio);  
	serio_set_drvdata(serio, NULL);  
	kfree(irtouch);  
	kfree(devContext);
}  
  
static void char_reg_setup_cdev(void)
{
	int error, devno = MKDEV(IRTOUCH_MAJOR_NUM,IRTOUCH_MINOR_NUM);

	cdev_init(&cdev,&ts_fops);
	cdev.owner = THIS_MODULE;
	cdev.ops = &ts_fops;
	error = cdev_add(&cdev,devno,1);

	if(error)
	{
		printk(KERN_ALERT "Adding char_reg_setup_cdev error=%d",error);
	}
}

/* 
 * irtouch_connect() is the routine that is called when someone adds a 
 * new serio device that supports Gunze protocol and registers it as 
 * an input device. 
 */  
static int irtouch_connect(struct serio *serio, struct serio_driver *drv)  
{  
	int err;
	struct input_dev *input_dev;
	
	if (!(irtouch = kmalloc(sizeof(struct irtouch), GFP_KERNEL)))  
	return -ENOMEM;  

	if (!(devContext = kmalloc(sizeof(struct device_context), GFP_KERNEL)))  
	return -ENOMEM; 

	memset(irtouch, 0, sizeof(struct irtouch)); 
	memset(devContext,0,sizeof(struct device_context));

	devContext->irtouchName = "IRTOUCH Serial TouchScreen";

	//create device node
	dev = MKDEV(IRTOUCH_MAJOR_NUM,IRTOUCH_MINOR_NUM);

	err = register_chrdev_region(dev,1,TOUCH_DEVICE_NAME);
	if(err < 0)
	{
		printk(KERN_ALERT "register chrdev error.\n");
		return err;
	}

	char_reg_setup_cdev();

	irser_class = class_create(THIS_MODULE, TOUCH_DEVICE_NAME);
	if(IS_ERR(irser_class))
	{
		printk(KERN_ALERT "class create failed.\n");
		return PTR_ERR(irser_class);
	}
	
	device_create(irser_class,NULL,MKDEV(IRTOUCH_MAJOR_NUM,0),NULL,TOUCH_DEVICE_NAME);


	input_dev = input_allocate_device();

	if (!irtouch || !input_dev){
		err = -ENOMEM;
		goto fail1;
	}

	irtouch->id = serio->id.id;  
	irtouch->serio = serio;  
	irtouch->dev = input_dev;

	irtouch->dev->evbit[0] = BIT(EV_KEY) | BIT(EV_ABS);
	set_bit(BTN_TOUCH, irtouch->dev->keybit);
	irtouch->dev->absbit[0] = BIT(ABS_X) | BIT(ABS_Y);
	set_bit(ABS_PRESSURE, irtouch->dev->absbit);

	//dual-touch
	input_set_abs_params(irtouch->dev, ABS_MT_POSITION_X, 0, 32767, 0, 0);
	input_set_abs_params(irtouch->dev, ABS_MT_POSITION_Y, 0, 32767, 0, 0);
	input_set_abs_params(irtouch->dev, ABS_MT_TOUCH_MAJOR, 0, 32767, 0, 0);
	input_set_abs_params(irtouch->dev, ABS_MT_TOUCH_MINOR, 0, 32767, 0, 0);
	input_set_abs_params(irtouch->dev, ABS_MT_TRACKING_ID, 0, 10, 0, 0);

	sprintf(irtouch->phys, "%s/input0", serio->phys);  

	irtouch->dev->name = devContext->irtouchName;  
	irtouch->dev->phys = irtouch->phys;  
	irtouch->dev->id.bustype = BUS_RS232;  
	irtouch->dev->id.vendor = 0x6615;//SERIO_UNKNOWN;  
	irtouch->dev->id.product = irtouch->id;  
	irtouch->dev->id.version = 0x0100;  

	serio_set_drvdata(serio, irtouch);  

	err = serio_open(serio, drv);  
	if (err) {  
		serio_set_drvdata(serio, NULL);  
		kfree(irtouch); 
		kfree(devContext); 
		return err;  
	}  

	irtouch_get_device_param(serio);

	err = input_register_device(irtouch->dev);  
	if(err)
	{
		goto fail1;
	}

	printk(KERN_INFO "input: %s on %s\n", devContext->irtouchName, serio->phys);  
	
	return 0;  

fail1:	input_free_device(input_dev);
	kfree(irtouch);
	kfree(devContext);
	return err;
}  
  
/* 
 * The serio driver structure. 
 */  
  
static struct serio_device_id irtouch_serio_ids[] = {  
	{  
		.type   = SERIO_RS232,  
		.proto  = SERIO_ANY,  
		.id = SERIO_ANY,  
		.extra  = SERIO_ANY,  
	},  
	{ 0 }  
};  
  
MODULE_DEVICE_TABLE(serio, irtouch_serio_ids);  
  

static struct serio_driver irtouch_drv = {  
	.driver     = {  
		.name   = "irtouchser",  
	},  
	.description    = DRIVER_DESC,  
	.id_table   = irtouch_serio_ids,  
	.interrupt  = irtouch_interrupt,  
	.connect    = irtouch_connect,  
	.disconnect = irtouch_disconnect,  
};  
  
/* 
 * The functions for inserting/removing us as a module. 
 */  
static int __init irtouch_init(void)  
{  
	int ret = -1;
	
	ret = serio_register_driver(&irtouch_drv); 
	if(ret == 0)
	{
		printk(KERN_INFO KBUILD_MODNAME ": " DRIVER_VERSION ":" DRIVER_DESC);
	} 

	return 0;  
}  
  
static void __exit irtouch_exit(void)  
{  
	serio_unregister_driver(&irtouch_drv);  
}  
  
module_init(irtouch_init);  
module_exit(irtouch_exit); 


